/*
    Copyright (c) 2002 Jens Tinz
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/**
 * @fileoverview Module for code used by the options dialog (options.xul).
 */

const menu = Components.utils.import(
    'resource://radialcontext-mz/menu.js', null);
const actions = Components.utils.import(
    'resource://radialcontext-mz/actions.js', null);
const utils = Components.utils.import(
    'resource://radialcontext-mz/utils.js', null);

let (service = Components.classes['@mozilla.org/intl/stringbundle;1']
     .getService(Components.interfaces.nsIStringBundleService)) {;
    const mainBundle = service.createBundle(
        'chrome://radialcontext-mz/locale/options.properties');
    const menusBundle = service.createBundle(
        'chrome://radialcontext-mz/locale/menus.properties');
    const categoriesBundle = service.createBundle(
        'chrome://radialcontext-mz/locale/categories.properties');
}

const XULNS = 'http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul';

const prefbranch = Components.classes["@mozilla.org/preferences-service;1"]
    .getService(Components.interfaces.nsIPrefService)
    .getBranch('extensions.radialcontext-mz.')
    .QueryInterface(Components.interfaces.nsIPrefBranch2);


function _addTreeCell(row, label, value) {
    let cell = row.ownerDocument.createElementNS(XULNS, 'treecell');
    row.appendChild(cell);
    if (label !== null)
        cell.setAttribute('label', label);
    // We later access this through treeview.getCellValue.
    // The description for that method in nsITreeView.idl is strange
    // ("This method is only called for columns of type other than |text|."?)
    // but it seems to behave as expected.
    if (value !== null)
        cell.setAttribute('value', value);
}

/**
 * Takes a treeitem node, removes it, removes the parent if now empty.
 *
 * This is currently not recursive.
 */
function _removeTreeLeaf(treeitem) {
    let treechildren = treeitem.parentNode;
    treechildren.removeChild(treeitem);
    if (!treechildren.hasChildNodes()) {
        let parentitem = treechildren.parentNode;
        parentitem.parentNode.removeChild(parentitem);
    }
}

function Pane(pane) {
    this.pane = pane;
}

Pane.prototype = {
    handleEvent: utils.handleEvent,
    shutdown: function() {},
    accept: function() {},

    getElementById: function(id) {
        return this.pane.ownerDocument.getElementById(id);
    },

    _setEnabled: function(id, enabled) {
        if (enabled)
            this.getElementById(id).removeAttribute('disabled');
        else
            this.getElementById(id).setAttribute('disabled', 'true');
    },

    get instantApply() {
        return this.pane.parentNode.instantApply;
    }
};


function BehaviorPane(pane) {
    Pane.call(this, pane);
    for each (let prefid in this.prefids) {
        let pref = this.getElementById(prefid);
        pref.addEventListener('change', this, false);
        // HACK: fire a changed event now to initialize:
        pref.preferences.fireChangedEvent(pref);
    }
}

BehaviorPane.prototype = {
    __proto__: Pane.prototype,

    prefids: [
        'rc-highlight-enabled-pref',
    ],

    on_change: function(event) {
        switch (event.originalTarget.id) {
        case 'rc-highlight-enabled-pref':
            this._setEnabled('highlight-broadcaster',
                             event.originalTarget.value);
            break;
        }
    },

    shutdown: function() {
        for each (let prefid in this.prefids)
            this.getElementById(prefid).removeEventListener('change', this,
                                                            false);
    }
};


function TooltipsPane(pane) {
    Pane.call(this, pane);

    // Duplicated from BehaviorPane, refactor if duplication increases.
    let pref = this.getElementById('rc-ttip-enabled-pref');
    pref.addEventListener('change', this, false);
    // HACK: fire a changed event now to initialize:
    pref.preferences.fireChangedEvent(pref);
}

TooltipsPane.prototype = {
    __proto__: Pane.prototype,

    on_change: function(event) {
        this._setEnabled('ttip-enabled-broadcaster',
                         event.originalTarget.value);
    },

    shutdown: function() {
        this.getElementById('rc-ttip-enabled-pref')
            .removeEventListener('change', this, false);
    }
};


// The plan: we want to support non-instantApply, so we need the
// actions code to ignore the pref settings and apply our own current
// customizations by hand. We store those as dom nodes in the xul
// tree element we use. When the user hits apply/close we flush
// stuff to prefs.
//
// To support instantApply we can do at least two things:
// - implement nsITreeView to tie the tree to the pref system
// - use a normal xul tree manually synced through a pref observer
//
// Because nsITreeView is very much not fun to implement (it is not
// purely a data model thing like the gtk treeview has, there is
// a bunch of bookkeeping to implement) we go for route two.
// When the user edits something we instantapply it to the pref
// system, which triggers our pref observer which updates the tree.
//
// We also need to store the uncustomized prefs somewhere. Currently we
// just keep this.manager.layouts vanilla and apply customizations
// before passing the layout to PieMenu.
//
// I am not 100% sure the ui as currently implemented will make sense
// to new users. The idea is the selected item in the treeview and
// the values in the menulists are synced. If the selection changes
// the menus are updated. If the menu values change and there is a
// selection the selected item is either updated or unselected.
// If the menu values change and there is no selection that remains
// the case (so randomly poking around in the menus will not select
// and then update something in the tree).
//
// the menuPreview shows the menu from the "menu" menulist with
// the customizations from the tree.
//
// The problem is deciding when to modify the selected item and when
// to unselect it. Currently only changing the customized action updates.
// Making the rest also update makes it easy to accidentally overwrite
// a different customization. Downside is moving an item requires an
// add/delete cycle (but it should be fairly easy to figure out:
// - select current item
// - modify "entry"
// - hey, nothing happened (except for selection vanishing), lets hit add
// - hey, now I have two, let's select the other and hit remove
// )
function CustomizePane(pane) {
    Pane.call(this, pane);

    this.manager = new actions.ActionManager(null);

    // Initialize tree-related bits:
    let tree = this.getElementById('customize-tree');
    this.treeview = tree.view;
    this.treebody = tree.body;
    this.locationColumn = tree.columns.getNamedColumn(
        'customize-tree-location');
    this.customizedColumn = tree.columns.getNamedColumn(
        'customize-tree-customized');
    tree.addEventListener('select', this, false);

    // Initialize the list of customizable menus:
    let menulist = this.getElementById('customize-menu-menu');
    let layoutNames = [name for (name in this.manager.layouts)];
    layoutNames.sort();

    for each (let name in layoutNames)
        menulist.appendItem(menusBundle.GetStringFromName(name), name);

    // "select" is triggered when we change the value too.
    menulist.addEventListener('select', this, false);
    menulist.addEventListener('command', this, false);
    // This triggers on_select to initializes this.layout.
    // That function looks at this.menuPreview, so set that:
    this.menuPreview = null;
    menulist.selectedIndex = 0;

    // Initialize the menu preview:
    let demoDiv = this.getElementById('demo');
    this.menuPreview = new menu.PieMenu(
        new menu.Theme(), demoDiv, this.layout, 170, 70,
        this.getElementById('customize-altbox').checked, 1, true, null);
    // Make the menu node a normal relatively-positioned node centered
    // in the preview div (instead of a fixed-positioned one):
    let style = this.menuPreview.node.style;
    style.position = 'relative';
    style.left = '50%';
    style.top = '50%';
    style.zIndex = '0';
    this.menuPreview.showLabels();
    // TODO: it would be neat to make the size of the menu label grow
    // the preview div (so we never clip) but I think that requires
    // some rearranging of the menu's nodes. Specifically currently the
    // main label node is sized by the center image, not the entire label.
    // If we really want to do this we probably need to add this mode
    // to the actual menu code, doing it here would require too much
    // prodding at the guts of the menu.

    // Hook up an event handler to the menu entry/position menulist:
    menulist = this.getElementById('customize-menu-entry');
    menulist.addEventListener('command', this, false);

    // Initialize the action selection menu:
    let menupopup = this.getElementById('customize-action-popup');
    // Specialcase the "dummy" item because we need to fix up its
    // label and because it is the only one not in any category.
    let dummyItem = menupopup.appendChild(
        this.pane.ownerDocument.createElementNS(XULNS, 'menuitem'));
    dummyItem.value = 'dummy';
    // Apparently an attribute but not a property:
    dummyItem.setAttribute('label', mainBundle.GetStringFromName('dummy'));
    // XXX the menuitem should be non-iconic, but I canont figure out
    // how to do the equivalent to the submenu entries.
    // So leave this out for now.
    // dummyItem.className = 'menuitem-non-iconic';
    this.setSelectedAction('dummy');
    let actionNames = [name for (name in this.manager.actions)
                            if (name !== 'dummy')];
    // This sorts by category as long as everything actually has a category.
    actionNames.sort();
    let lastCategoryName = null;
    let categoryPopup;
    for each (let actionName in actionNames) {
        let [categoryName, rest] = actionName.split('_', 2);
        if (categoryName !== lastCategoryName) {
            let submenu = menupopup.appendChild(
                this.pane.ownerDocument.createElementNS(XULNS, 'menu'));
            // label is an attribute but not a property:
            submenu.setAttribute(
                'label', categoriesBundle.GetStringFromName(categoryName));
            categoryPopup = submenu.appendChild(
                this.pane.ownerDocument.createElementNS(XULNS, 'menupopup'));
            lastCategoryName = categoryName;
        }
        let action = this.manager.actions[actionName];
        let menuitem = categoryPopup.appendChild(
            this.pane.ownerDocument.createElementNS(XULNS, 'menuitem'));
        // Apparently an attribute but not a property.
        menuitem.setAttribute('label', action.label);
        menuitem.setAttribute(
            'image', this.menuPreview.theme.getImage(action.iconname));
        menuitem.value = actionName;
        menuitem.className = 'menuitem-iconic';
    }
    menupopup.addEventListener('command', this, false);

    // Read customizations from the pref system into the treeview
    // by firing our pref observer for all set prefs.
    // (We could do something fancier to avoid the parent lookup
    // in observe() but there is no real point.)
    for each (let key in prefbranch.getChildList('customizations.', {}))
        this.observe(prefbranch, 'nsPref:changed', key);

    // Hook up final event handlers and observers:

    // Assume this.instantApply can change between here and shutdown,
    // so remember whether we need to unregister our observer:
    this.observerAdded = this.instantApply;
    if (this.observerAdded)
        prefbranch.addObserver('customizations.', this, false);

    for each (let id in ['customize-add',
                         'customize-remove',
                         'customize-altbox',
                        ])
        this.getElementById(id).addEventListener('command', this, false);
}

CustomizePane.prototype = {
    __proto__: Pane.prototype,

    shutdown: function() {
        if (this.observerAdded)
            prefbranch.removeObserver('customizations.', this);

        for each (let [id, event] in [
                ['customize-tree', 'select'],
                ['customize-menu-menu', 'select'],
                ['customize-menu-menu', 'command'],
                ['customize-menu-entry', 'command'],
                ['customize-action-popup', 'command'],
                ['customize-add', 'command'],
                ['customize-remove', 'command'],
            ])
            this.getElementById(id)
                .removeEventListener(event, this, false);
    },

    observe: function(prefbranch_ignored, topic, key) {
        if (topic !== 'nsPref:changed')
            throw new Error('huh, I did not sign up for ' + topic);

        let entry = this.manager.parseKey(prefbranch, key, true);
        if (entry === null)
            return;

        let [layoutName, index, actionName] = entry;
        if (actionName === null) {
            let layoutIndex = this._findLayout(layoutName);
            let menulist = this.getElementById('customize-menu-menu');
            // If layoutIndex is -1 (should not happen) this runs 0 times.
            for (let i in this._findLayoutChildren(layoutIndex)) {
                if (this.treeview.getCellValue(i, this.locationColumn)
                    === String(index)) {
                    // Found it in the tree.
                    if (menulist.value === layoutName)
                        // We are currently previewing this menu.
                        this._revertLayout(layoutName, index);

                    _removeTreeLeaf(this.treeview.getItemAtIndex(i));
                    break;
                }
            }
        } else
            this._setCustomization(layoutName, index, actionName);
    },

    accept: function() {
        if (!this.instantApply) {
            // This should be resetBranch but that returns "not implemented".
            // Then I tried deleteBranch but apparently that does not fire
            // pref observers. So do it by hand:
            for each (let key in prefbranch.getChildList('customizations.',
                                                         {}))
                prefbranch.clearUserPref(key);

            prefbranch.deleteBranch('customizations.');
            let layoutName = null;
            for (let i = 0; i < this.treeview.rowCount; ++i) {
                if (!this.treeview.getLevel(i))
                    layoutName =
                        this.treeview.getCellValue(i, this.locationColumn);
                else
                    this._setPref(
                        layoutName,
                        this.treeview.getCellValue(i, this.locationColumn),
                        this.treeview.getCellValue(i, this.customizedColumn));
            }
        }
    },

    /**
     * Convert an action name ('menu_tab') to a fancy string ('Menu -> Tab').
     */
    actionNameToDescription: function(name) {
        // HACK: specialcase this (because we do not want this label to
        // show up in the menu preview):
        if (name === 'dummy')
            return mainBundle.GetStringFromName('dummy');
        let label = this.manager.actions[name].label || name;
        let bits = name.split('_', 2);
        if (bits.length === 1)
            return label;

        let categoryLabel = categoriesBundle.GetStringFromName(bits[0]);
        return mainBundle.formatStringFromName('tree_category_action',
                                               [categoryLabel, label], 2);
    },

    /**
     * Update the selected action in the menubutton.
     */
    setSelectedAction: function(name) {
        // I experimented with putting the category above the action name
        // in the menu button, but I think it looked a bit confusing
        // (especially since I saw no way to squeeze a better indication
        // of "this is the category/container" vs "this is the child"
        // than a tiny arrow in there). And there is definitely not
        // enough room to put the category before the action name :(
        this.selectedAction = name;
        // We need to use the dummy image instead of no image to
        // get the size of the button right (or it will resize when
        // we select a non-dummy).
        // XXX Would be better if this was vertical-only
        // (make the "Blank (remove item)" align with the "Top right"
        // in the menulist above it instead of shifting it to the right
        // with the dummy image).
        // Perhaps we should play games with css min height and
        // the theme's size props instead of using the image?
        let iconname = 'dummy';
        let category = '';
        let label = mainBundle.GetStringFromName('dummy');
        // HACK: specialcase the dummy like actionNameToDescription does:
        if (name !== 'dummy') {
            let action = this.manager.actions[name];
            label = action.label;
            iconname = action.iconname;
            let bits = name.split('_', 2);
            if (bits.length === 2)
                category = categoriesBundle.GetStringFromName(bits[0]);
        }
        let iconurl = this.menuPreview.theme.getImage(iconname);
        this.getElementById('customize-action-image')
            .setAttribute('src', iconurl);
        this.getElementById('customize-action-description')
            .setAttribute('value', label);
    },

    _applyLayout: function(actionName, index) {
        this.layout[index] = this.manager.actions[actionName];
        this.menuPreview.applyLayout(this.layout);
    },

    _revertLayout: function(layoutName, index) {
        this.layout[index] = this.manager.layouts[layoutName][index];
        this.menuPreview.applyLayout(this.layout);
    },

    /**
     * Add a customized action to the tree and possibly the menu preview.
     * @param {str} layoutName Name of the layout to customize
     * @param {int} entry Position in the layout array to customize
     * @param {str} actionName Name of the action to put there
     */
    _setCustomization: function(layoutName, entry, actionName) {
        // This is called from the pref observer in instantapply mode
        // and when hitting the "add" button in non-instantapply mode.
        // Also called through the pref observer function when initializing
        // (manually, also in non-instantapply mode).

        // Update the menu preview if it is showing the modified layout:
        if (this.getElementById('customize-menu-menu').value === layoutName)
            this._applyLayout(actionName, entry);

        let layoutIndex = this._findLayout(layoutName);
        let entryIndex = -1;
        for (let i in this._findLayoutChildren(layoutIndex)) {
            if (parseInt(this.treeview.getCellValue(
                             i, this.locationColumn)) === entry) {
                // There is an existing entry, update it:
                this.treeview.setCellText(
                    i, this.customizedColumn,
                    this.actionNameToDescription(actionName));
                this.treeview.setCellValue(
                    i, this.customizedColumn, actionName);
                entryIndex = i;
                break;
            }
        }
        if (entryIndex === -1) {
            // There was no existing entry, create one.
            // First create the parent node if it does not exist yet:
            let children;
            let doc = this.treebody.ownerDocument;
            if (layoutIndex === -1) {
                let item = this.treebody.appendChild(
                    doc.createElementNS(XULNS, 'treeitem'));
                item.setAttribute('container', 'true');
                item.setAttribute('open', 'true');
                let row = doc.createElementNS(XULNS, 'treerow');
                // Name as displayed in the first column:
                _addTreeCell(row, menusBundle.GetStringFromName(layoutName),
                             layoutName);
                // "original" and "customized" are empty:
                _addTreeCell(row, null, null);
                _addTreeCell(row, null, null);
                item.appendChild(row);
                children = item.appendChild(
                    doc.createElementNS(XULNS, 'treechildren'));
            } else
                // It exists, so it has at least one child, which we can abuse
                // to get at its "children" node:
                children = this.treeview.getItemAtIndex(
                    layoutIndex + 1).parentNode;

            let entries = this.getElementById(
                'customize-menu-entry').menupopup.childNodes;
            let item = children.appendChild(
                doc.createElementNS(XULNS, 'treeitem'));
            let row = item.appendChild(doc.createElementNS(XULNS, 'treerow'));

            _addTreeCell(row, entries[entry].getAttribute('label'),
                         String(entry));
            let origAction = this.manager.layouts[layoutName][entry];
            _addTreeCell(row, this.actionNameToDescription(origAction.name),
                         null);
            _addTreeCell(row, this.actionNameToDescription(actionName),
                         actionName);

            entryIndex = this.treeview.getIndexOfItem(item);
        }
        // Update tree selection if there is none.
        // Notice this fires a select event, so we should do it after
        // updating the values in the tree or that event handler will
        // modify the menu values.
        if (this.treeview.selection.currentIndex === -1 ||
            this.treeview.getParentIndex(
                this.treeview.selection.currentIndex) === -1)
            this.treeview.selection.select(entryIndex);
    },

    on_select: function(event) {
        switch (event.currentTarget.id) {
        case 'customize-menu-menu':
            // Update the menu preview to match the selected menu.
            // Make a copy of the layout since we will apply customizations:
            let value = event.currentTarget.value;
            let layout = this.manager.layouts[value].slice(0);

            for (let i in this._findLayoutChildren(this._findLayout(value))) {
                let index = parseInt(
                    this.treeview.getCellValue(i, this.locationColumn));
                layout[index] = this.manager.actions[
                    this.treeview.getCellValue(i, this.customizedColumn)];
            }

            this.layout = layout;

            // This gets called by the constructor to initialize
            // this.layout before this.menuPreview exists, so check:
            if (this.menuPreview !== null)
                this.menuPreview.applyLayout(layout);

            break;
        case 'customize-tree':
            let addButton = this.getElementById('customize-add');
            let removeButton = this.getElementById('customize-remove');
            let index = this.treeview.selection.currentIndex;
            if (index === -1) {
                addButton.disabled = false;
                removeButton.disabled = true;
                return;
            }
            // Tree selection changed: Update menus
            // (which may update the preview).
            let menulist = this.getElementById('customize-menu-menu');
            if (this.treeview.isContainer(index)) {
                // menu name selected
                menulist.value =
                    this.treeview.getCellValue(index, this.locationColumn);
                addButton.disabled = false;
                removeButton.disabled = true;
            } else {
                addButton.disabled = true;
                removeButton.disabled = false;
                // menu entry selected
                menulist.value = this.treeview.getCellValue(
                    this.treeview.getParentIndex(index), this.locationColumn);

                menulist = this.getElementById('customize-menu-entry');
                menulist.selectedIndex = parseInt(
                    this.treeview.getCellValue(index, this.locationColumn));

                this.setSelectedAction(
                    this.treeview.getCellValue(index, this.customizedColumn));
            }
            break;
        default:
            throw new Error(
                'RadialContext: unhandled select for ' +
                event.target.localName + ' ' + event.target.id);
        }
    },

    on_command: function(event) {
        let actionName, index, parent, layoutName;
        // (for menulists this is fired on the item, apparently, so
        // we really need currentTarget here, not target).
        switch (event.currentTarget.id) {
        case 'customize-menu-menu':
        case 'customize-menu-entry':
            // Lose the selection in the tree if there is one. (could
            // change the menu the current item is in but that is
            // probably unwanted, especially since it could overwrite
            // something else)

            // Beware: the documentation
            // http://developer.mozilla.org/en/docs/
            // XUL_Tutorial:Tree_Selection
            // is slightly misleading here: although it mentions you
            // should not assign to currentIndex to change the
            // selection, it does not say why or how to properly
            // unselect stuff. The reason is hinted at in
            // nsITreeSelection.idl: currentIndex is not the
            // *selected* row, it is the *current* or *focused* row
            // (the one with a focus rect around it). *Normally* that
            // row is also selected (and vice versa), but this is not
            // guaranteed. Specifically clearSelection does not touch
            // currentIndex! Result:
            // - Tree has a row selected
            // - User changes menuentry (this function runs)
            // - This calls clearSelection
            // - clearSelection visually clears the selection
            //   but leaves currentIndex
            // - clearSelection fires a select event (on_select runs)
            // - that uses currentIndex to set this menulist back to what it
            //   thinks is the currently selected layout
            // so it looks like the menulist breaks. Conclusion: in
            // single-select mode *only* use the select method to
            // manipulate the selection, since that one updates both
            // currentIndex and the selection.
            this.treeview.selection.select(-1);
            break;
        case 'customize-action-popup':
            // Make the button show the selected action:
            this.setSelectedAction(event.originalTarget.value);
            // Update the selected item if there is one.
            let treeIndex = this.treeview.selection.currentIndex;
            if (treeIndex === -1)
                return;

            parent = this.treeview.getParentIndex(treeIndex);
            if (parent === -1)
                return;

            let entry =
                this.treeview.getCellValue(treeIndex, this.locationColumn);

            if (this.instantApply)
                this._setPref(
                    this.treeview.getCellValue(parent, this.locationColumn),
                    entry, this.selectedAction);
            else {
                // Update the tree and preview:
                this.treeview.setCellValue(
                    treeIndex, this.customizedColumn, this.selectedAction);
                this.treeview.setCellText(
                    treeIndex, this.customizedColumn,
                    this.actionNameToDescription(this.selectedAction));
                this._applyLayout(this.selectedAction, parseInt(entry));
            }

            break;
        case 'customize-add':
            // This should be effectively a noop if there is a selection.
            layoutName = this.getElementById('customize-menu-menu').value;
            index = this.getElementById('customize-menu-entry')
                .selectedIndex;
            if (this.instantApply)
                this._setPref(layoutName, index, this.selectedAction);
            else
                this._setCustomization(layoutName, index,
                                       this.selectedAction);
            break;
        case 'customize-remove':
            index = this.treeview.selection.currentIndex;

            // Ignore attempts to remove without an item selected.
            if (index === -1)
                return;

            // TODO: maybe support resetting an entire menu this way?
            parent = this.treeview.getParentIndex(index);
            if (parent === -1)
                return;

            layoutName =
                this.treeview.getCellValue(parent, this.locationColumn);
            let layoutIndex =
                this.treeview.getCellValue(index, this.locationColumn);

            if (this.instantApply)
                prefbranch.clearUserPref(
                    'customizations.' + layoutName + '.' +
                    String(layoutIndex));
            else {
                this._revertLayout(layoutName, layoutIndex);
                _removeTreeLeaf(this.treeview.getItemAtIndex(index));
            }
            break;
        case 'customize-altbox':
            this.menuPreview.updateEntries(event.currentTarget.checked);
            break;
        default:
            throw new Error(
                'RadialContext: unhandled command for ' +
                event.target.localName + ' ' + event.target.id);
        }
    },

    /**
     * Helper returning the index of the node for a layout name or -1.
     */
    _findLayout: function(layoutName) {
        if (!this.treeview.rowCount)
            return -1;
        let index = 0;
        while (true) {
            if (this.treeview.getCellValue(index, this.locationColumn)
                === layoutName)
                return index;
            // I do not see a "next sibling" method in nsITreeView so
            // take a detour through the DOM:
            let next = this.treeview.getItemAtIndex(index).nextSibling;
            if (next === null)
                return -1;
            index = this.treeview.getIndexOfItem(next);
        }
    },

    /**
     * Generator yielding tree indices for a layout.
     *
     * @param {int} index The result from _findLayout.
     */
    _findLayoutChildren: function(index) {
        if (index === -1)
            return;
        for (++index;
             index < this.treeview.rowCount && this.treeview.getLevel(index);
             ++index)
            yield index;
    },

    _setPref: function(layoutName, index, actionName) {
        prefbranch.setCharPref(
            'customizations.' + layoutName + '.' + String(index), actionName);
    }
};


/**
 * @class Wrapper for a prefwindow, forwarding events to Pane instances.
 *
 * Only used as event listener, nothing else should get a reference.
 */
function OptionsWindow(window) {
    this.window = window;
    this.panes = [];
    window.addEventListener('dialogaccept', this, false);
    window.addEventListener('dialogcancel', this, false);
    window.addEventListener('paneload', this, false);
}

OptionsWindow.prototype = {
    handleEvent: utils.handleEvent,

    on_dialogaccept: function(event) {
        for each (let pane in this.panes)
            pane.accept();

        this.shutdown();
    },

    on_dialogcancel: function(event) {
        // Perhaps we should forward this to the panes too,
        // but let's add that feature when we actually need it.
        this.shutdown();
    },

    shutdown: function() {
        for each (let pane in this.panes)
            pane.shutdown();

        this.window.removeEventListener('dialogaccept', this, false);
        this.window.removeEventListener('dialogcancel', this, false);
        this.window.removeEventListener('paneload', this, false);
    },

    on_paneload: function(event) {
        let klass = this.paneClasses[event.target.id];
        if (klass !== undefined)
            this.panes.push(new klass(event.target));
    },

    paneClasses: {
        'pane-behavior': BehaviorPane,
        'pane-tooltips': TooltipsPane,
        'pane-customize': CustomizePane
    }
};

/**
 * Call "new OptionsWindow(window)". Helper used from a script tag in xul.
 */
function bind(window) {
    new OptionsWindow(window);
}
